//----------------------------------------------------------------------------------
// PSK31.C
//
// PSK31 beacon (PIC16F872 @ 16Mhz)
// 16Mhz give me exact timing, I have tried 20Mhz crystal but timer resolution
// is not enough to have precise baudrate. (31.25baud is very critical)
//
// Created by : Remi Bilodeau VE2YAG 31 mars 2002
// Last Modification : 11 avril 2002 Version 1.1 built 01
//
// CPU power is not enough to generate PSK31 modulation. To correct that I have
// digitalized a 1khz carrier, with amplitude modulation at 31.25hz sinus. The
// modulation is digitalized at 16khz. I have tried 64khz frequency but CPU is
// too slow and 16khz is more than enough to have good waveform.
//
// 1- For 1khz carrier full wave cycle I have 16 sample to output to R-R2 network.
// 2- The modulation cycle is 32ms (31.25baud), 32 time the carrier:
//    16 samples x 32 cycle = 512 total sample.
//
// 512 sample(bytes) is too much for small F872 memory, I have to cut a little bit.
//
// 1-Carrier cycle, negative side is a copy of positive, I have erased
//   all negative side of carrier (180 to 360 degre) cut memory to 256 bytes.
// 2-Modulation, rising side (0% to 100% modulation) is a copy of falling
//   side, falling side erased.
//
// Sample table is down to 128 bytes and just a small CPU power is necessary
// to restore full modulation waveform. (in interrupt function)
//
// 16 samples = 1khz sinus, sythetised at 16khz
// 1000us     = 1khz period,
// CPU Tcycle = 0.25us@16Mhz (Tosc*4)
//
// (1000us / 16samples) / Tcycle = ?
//     (62.5us)         / 0.25us  = 250 cycles
//
// 250 cycle between interrupt: .25us * 250 cycles * 16 samples  = 1000.0Hz frequency tone
//
//
// Usage:
// Type any key to enter in config mode. The beacon is shutdown in this mode. Type
// R to restore beacon activity.
//
// B<text>    PSK31 beacon text. Type ^ to switch to CW. Type ~ to output
//            1 sec of 1khz tone. (for CW beacon or end of PSK31 modulation)
//
//              Example: BT VE2YAG Beacon, FN19ES k k k~^  DE VE2YAG
//                       BT ^VE2YAG FN19 ~~ (CW only beacon)
//              note: Use space in CW to insert small delay betwen PSK and CW.
// E<00-FF>     Wait x second before transmission, 00=countinous sending, in hexadecimal.
// R            Reset beacon, and start sending.
// D            Send in hex: CPUTYPE + SOFTVERSION + EEPROMSIZE + 00 + (dump en eeprom)
//
//
// History:
// 03/31/2002  V1.0b01  Initial version for PIC16F872.
// 04/11/2002  V1.1     bug: PTT not release when enter config mode
//
// Pinout:
// TMR0 Alan and AX25 timing
//
// RA0
// RA1
// RA2
// RA3
// RA4
// RA5
// RB0 R-R2 DAC output 1
// RB1 R-R2 DAC output 2
// RB2 R-R2 DAC output 3
// RB3 R-R2 DAC output 4
// RB4 R-R2 DAC output 5
// RB5 R-R2 DAC output 6
// RB6 R-R2 DAC output 7
// RB7 R-R2 DAC output 8
// RC0
// RC1
// RC2
// RC3
// RC4
// RC5
// RC6 Serial port Output
// RC7 Serial port Input
//
//-------------------------------------------------
#include <pic.h>
#include <ctype.h>

#define CFG_CPUTYPE     0x72  // Type of CPU 16F872
#define CFG_SOFTVERSION 0x10  // Version 1.0
#define CFG_EEPROMSIZE  64    // Size of eeprom

/*{{{  Define*/


#define FALSE 	0           // Value of False
#define TRUE	!FALSE      // Value of True

#define	PORTBIT(adr, bit)	((unsigned)(&adr)*8+(bit))
#define CLEARWATCHDOG() asm("clrwdt")

#define byte unsigned char
#define word unsigned int

#define WaitCycleEnd() while(ModCount!=0) if(RC7==0) break
#define WaitCycleStart() while(ModCount!=1) if(RC7==0) break

/*}}}  */

/*{{{  variable*/

const unsigned char hex[] = {"0123456789ABCDEF"};

#define MODE_PSK 0
#define MODE_CW 1

#define PTT RC4     // Radio PTT, active +5v
#define KEY RC5     // CW Key, active +5v
#define SERIN RC7   // Serial in (modify in sbit function also)

//-----------------------------------------------------
// EEPROM map
//-----------------------------------------------------
#define MAXCARACTER   49    // Maximum caracter in line buffer (Beacon size-1)

#define EE_BEACONDELAY    0
#define EE_BEACONTEXTSIZE 1
#define EE_BEACONTEXT     16

//-----------------------------------------------------
// Modulator variable
//-----------------------------------------------------
#define FLAT     0
#define RISING   1
#define FALLING  2
#define NONE     3
unsigned char ModType;    // type of modulation for half-cycle
unsigned char ModCount;   // 0-255 count for half-cycle
unsigned char ModPhase;   // 0 or 1 for audio signal phase reversal


//-----------------------------------------------------
// Config mode variable
//-----------------------------------------------------
#define BK             8       /* Backspace key */
#define CR             0x0D    /* Carriage return key */
#define DEL            127     /* Delete key */
#define CTRL_C         0x03    /* Ctrl-C key */
unsigned char LineCount;                  // Number of caracter in buffer
unsigned char LineBuffer[MAXCARACTER];    // Line buffer
const    byte bk_str[] = { BK,' ',BK,0 }; // Backspace string


//----------------------------------------------------------
// Digitalized PSK31 modulation.
// sinus at 1 khz, only rising side, positive carrier only
// 8 samples(positive side, carrier)
// 16 cycles from 100% to 0% modulated signal
// 8 * 16 = 128 total samples
//----------------------------------------------------------
const unsigned char Waveform[] = { \
        128, 175, 217, 245, 254, 245, 217, 175,  \
        128, 175, 216, 244, 253, 244, 216, 175,  \
        128, 174, 214, 240, 248, 240, 213, 173,  \
        128, 171, 209, 234, 242, 233, 208, 170,  \
        128, 168, 204, 227, 234, 226, 202, 167,  \
        128, 165, 197, 217, 224, 216, 195, 163,  \
        128, 160, 189, 207, 212, 206, 187, 159,  \
        128, 156, 180, 196, 200, 194, 178, 154,  \
        128, 151, 171, 184, 188, 183, 169, 149,  \
        128, 147, 163, 173, 176, 171, 160, 145,  \
        128, 142, 154, 162, 164, 161, 152, 140,  \
        128, 138, 147, 152, 153, 151, 145, 137,  \
        128, 134, 140, 143, 144, 142, 138, 133,  \
        128, 131, 134, 136, 137, 136, 133, 130,  \
        128, 129, 131, 131, 131, 131, 130, 129,  \
        128, 128, 128, 128, 128, 128, 128, 128 };

//----------------------------------------------------------
// Varicode: The alphabet of PSK31, only 0-127 value.
//----------------------------------------------------------
const unsigned int Varicode[128] = {
	0xAAC0,	// ASCII =   0	1010101011
	0xB6C0,	// ASCII =   1	1011011011
	0xBB40,	// ASCII =   2	1011101101
	0xDDC0,	// ASCII =   3	1101110111
	0xBAC0,	// ASCII =   4	1011101011
	0xD7C0,	// ASCII =   5	1101011111
	0xBBC0,	// ASCII =   6	1011101111
	0xBF40,	// ASCII =   7	1011111101
	0xBFC0,	// ASCII =   8	1011111111
	0xEF00,	// ASCII =   9	11101111
	0xE800,	// ASCII =  10	11101
	0xDBC0,	// ASCII =  11	1101101111
	0xB740,	// ASCII =  12	1011011101
	0xF800,	// ASCII =  13	11111
	0xDD40,	// ASCII =  14	1101110101
	0xEAC0,	// ASCII =  15	1110101011
	0xBDC0,	// ASCII =  16	1011110111
	0xBD40,	// ASCII =  17	1011110101
	0xEB40,	// ASCII =  18	1110101101
	0xEBC0,	// ASCII =  19	1110101111
	0xD6C0,	// ASCII =  20	1101011011
	0xDAC0,	// ASCII =  21	1101101011
	0xDB40,	// ASCII =  22	1101101101
	0xD5C0,	// ASCII =  23	1101010111
	0xDEC0,	// ASCII =  24	1101111011
	0xDF40,	// ASCII =  25	1101111101
	0xEDC0,	// ASCII =  26	1110110111
	0xD540,	// ASCII =  27	1101010101
	0xD740,	// ASCII =  28	1101011101
	0xEEC0,	// ASCII =  29	1110111011
	0xBEC0,	// ASCII =  30	1011111011
	0xDFC0,	// ASCII =  31	1101111111
	0x8000,	// ASCII = ' '	1
	0xFF80,	// ASCII = '!'	111111111
	0xAF80,	// ASCII = '"'	101011111
	0xFA80,	// ASCII = '#'	111110101
	0xED80,	// ASCII = '$'	111011011
	0xB540,	// ASCII = '%'	1011010101
	0xAEC0,	// ASCII = '&'	1010111011
	0xBF80,	// ASCII = '''	101111111
	0xFB00,	// ASCII = '('	11111011
	0xF700,	// ASCII = ')'	11110111
	0xB780,	// ASCII = '*'	101101111
	0xEF80,	// ASCII = '+'	111011111
	0xEA00,	// ASCII = ','	1110101
	0xD400,	// ASCII = '-'	110101
	0xAE00,	// ASCII = '.'	1010111
	0xD780,	// ASCII = '/'	110101111
	0xB700,	// ASCII = '0'	10110111
	0xBD00,	// ASCII = '1'	10111101
	0xED00,	// ASCII = '2'	11101101
	0xFF00,	// ASCII = '3'	11111111
	0xBB80,	// ASCII = '4'	101110111
	0xAD80,	// ASCII = '5'	101011011
	0xB580,	// ASCII = '6'	101101011
	0xD680,	// ASCII = '7'	110101101
	0xD580,	// ASCII = '8'	110101011
	0xDB80,	// ASCII = '9'	110110111
	0xF500,	// ASCII = ':'	11110101
	0xDE80,	// ASCII = ';'	110111101
	0xF680,	// ASCII = '<'	111101101
	0xAA00,	// ASCII = '='	1010101
	0xEB80,	// ASCII = '>'	111010111
	0xABC0, // ASCII = '?'	1010101111
	0xAF40,	// ASCII = '@'	1010111101
	0xFA00,	// ASCII = 'A'	1111101
	0xEB00,	// ASCII = 'B'	11101011
	0xAD00,	// ASCII = 'C'	10101101
	0xB500,	// ASCII = 'D'	10110101
	0xEE00,	// ASCII = 'E'	1110111
	0xDB00,	// ASCII = 'F'	11011011
	0xFD00,	// ASCII = 'G'	11111101
	0xAA80,	// ASCII = 'H'	101010101
	0xFE00,	// ASCII = 'I'	1111111
	0xFE80,	// ASCII = 'J'	111111101
	0xBE80,	// ASCII = 'K'	101111101
	0xD700,	// ASCII = 'L'	11010111
	0xBB00,	// ASCII = 'M'	10111011
	0xDD00,	// ASCII = 'N'	11011101
	0xAB00,	// ASCII = 'O'	10101011
	0xD500,	// ASCII = 'P'	11010101
	0xEE80,	// ASCII = 'Q'	111011101
	0xAF00,	// ASCII = 'R'	10101111
	0xDE00,	// ASCII = 'S'	1101111
	0xDA00,	// ASCII = 'T'	1101101
	0xAB80,	// ASCII = 'U'	101010111
	0xDA80,	// ASCII = 'V'	110110101
	0xAE80,	// ASCII = 'W'	101011101
	0xBA80,	// ASCII = 'X'	101110101
	0xBD80,	// ASCII = 'Y'	101111011
	0xAB40,	// ASCII = 'Z'	1010101101
	0xFB80,	// ASCII = '['	111110111
	0xF780,	// ASCII = '\'	111101111
	0xFD80,	// ASCII = ']'	111111011
	0xAFC0,	// ASCII = '^'	1010111111
	0xB680,	// ASCII = '_'	101101101
	0xB7C0,	// ASCII = '`'	1011011111
	0xB000,	// ASCII = 'a'	1011
	0xBE00,	// ASCII = 'b'	1011111
	0xBC00,	// ASCII = 'c'	101111
	0xB400,	// ASCII = 'd'	101101
	0xC000,	// ASCII = 'e'	11
	0xF400,	// ASCII = 'f'	111101
	0xB600,	// ASCII = 'g'	1011011
	0xAC00,	// ASCII = 'h'	101011
	0xD000,	// ASCII = 'i'	1101
	0xF580,	// ASCII = 'j'	111101011
	0xBF00,	// ASCII = 'k'	10111111
	0xD800,	// ASCII = 'l'	11011
	0xEC00,	// ASCII = 'm'	111011
	0xF000,	// ASCII = 'n'	1111
	0xE000,	// ASCII = 'o'	111
	0xFC00,	// ASCII = 'p'	111111
	0xDF80,	// ASCII = 'q'	110111111
	0xA800,	// ASCII = 'r'	10101
	0xB800,	// ASCII = 's'	10111
	0xA000,	// ASCII = 't'	101
	0xDC00,	// ASCII = 'u'	110111
	0xF600,	// ASCII = 'v'	1111011
	0xD600,	// ASCII = 'w'	1101011
	0xDF00,	// ASCII = 'x'	11011111
	0xBA00,	// ASCII = 'y'	1011101
	0xEA80,	// ASCII = 'z'	111010101
	0xADC0,	// ASCII = '{'	1010110111
	0xDD80,	// ASCII = '|'	110111011
	0xAD40,	// ASCII = '}'	1010110101
	0xB5C0,	// ASCII = '~'	1011010111
	0xED40	// ASCII = 127	1110110101
    };

//---------------------------------------------------------------------------
// The word in the CW table is divided into 8 groups of two bits starting
// at the msb side.  The two bits represent one of four possible states.
//  00 - end of character
//  01 - DOT
//  10 - DASH
//  11 - SPACE of two dot times
//---------------------------------------------------------------------------
const unsigned int morse[59] = {
	0xF000,		// 1111 1100 0000 0000b  ( 32)       WORD SPACE
	0x0000,		// 0000 0000 0000 0000b  ( 33)  !
	0x0000,		// 0000 0000 0000 0000b  ( 34)  "
	0x0000,		// 0000 0000 0000 0000b  ( 35)  #
	0x0000,		// 0000 0000 0000 0000b  ( 36)  $
	0x0000,		// 0000 0000 0000 0000b  ( 37)  %
	0x0000,		// 0000 0000 0000 0000b  ( 38)  &
	0x0000,		// 0000 0000 0000 0000b  ( 39)  '
	0x0000,		// 0000 0000 0000 0000b  ( 40)  (
	0x0000,		// 0000 0000 0000 0000b  ( 41)  )
	0x566C,		// 0101 0110 0110 1100b  ( 42)  *    ...-.-  SK
	0x6670,		// 0110 0110 0111 0000b  ( 43)  +    .-.-.   AR
	0xA5AC,		// 1010 0101 1010 1100b  ( 44)  ,    --..--
	0x0000,		// 0000 0000 0000 0000b  ( 45)  -
	0x666C,		// 0110 0110 0110 1100b  ( 46)  .    .-.-.-
	0x9670,		// 1001 0110 0111 0000b  ( 47)  /    -..-.
	0xAAB0,		// 1010 1010 1011 0000b  ( 48)  0    -----
	0x6AB0,		// 0110 1010 1011 0000b  ( 49)  1    .----
	0x5AB0,		// 0101 1010 1011 0000b  ( 50)  2    ..---
	0x56B0,		// 0101 0110 1011 0000b  ( 51)  3    ...--
	0x55B0,		// 0101 0101 1011 0000b  ( 52)  4    ....-
	0x5570,		// 0101 0101 0111 0000b  ( 53)  5    .....
	0x9570,		// 1001 0101 0111 0000b  ( 54)  6    -....
	0xA570,		// 1010 0101 0111 0000b  ( 55)  7    --...
	0xA970,		// 1010 1001 0111 0000b  ( 56)  8    ---..
	0xAA70,		// 1010 1010 0111 0000b  ( 57)  9    ----.
	0x0000,		// 0000 0000 0000 0000b  ( 58)  :
	0x0000,		// 0000 0000 0000 0000b  ( 59)  ;
	0x0000,		// 0000 0000 0000 0000b  ( 60)  <
	0x95B0,		// 1001 0101 1011 0000b  ( 61)  =    -...-   BT
	0x0000,		// 0000 0000 0000 0000b  ( 62)  >
	0x5A5C,		// 0101 1010 0101 1100b  ( 63)  ?    ..--..
	0x0000,		// 0000 0000 0000 0000b  ( 64)  @
	0x6C00,		// 0110 1100 0000 0000b  ( 65)  A    .-
	0x95C0,		// 1001 0101 1100 0000b  ( 66)  B    -...
	0x99C0,		// 1001 1001 1100 0000b  ( 67)  C    -.-.
	0x9700,		// 1001 0111 0000 0000b  ( 68)  D    -..
	0x7000,		// 0111 0000 0000 0000b  ( 69)  E    .
	0x59C0,		// 0101 1001 1100 0000b  ( 70)  F    ..-.
	0xA700,		// 1010 0111 0000 0000b  ( 71)  G    --.
	0x55C0,		// 0101 0101 1100 0000b  ( 72)  H    ....
	0x5C00,		// 0101 1100 0000 0000b  ( 73)  I    ..
	0x6AC0,		// 0110 1010 1100 0000b  ( 74)  J    .---
	0x9B00,		// 1001 1011 0000 0000b  ( 75)  K    -.-
	0x65C0,		// 0110 0101 1100 0000b  ( 76)  L    .-..
	0xAC00,		// 1010 1100 0000 0000b  ( 77)  M    --
	0x9C00,		// 1001 1100 0000 0000b  ( 78)  N    -.
	0xAB00,		// 1010 1011 0000 0000b  ( 79)  O    ---
	0x69C0,		// 0110 1001 1100 0000b  ( 80)  P    .--.
	0xA6C0,		// 1010 0110 1100 0000b  ( 81)  Q    --.-
	0x6700,		// 0110 0111 0000 0000b  ( 82)  R    .-.
	0x5700,		// 0101 0111 0000 0000b  ( 83)  S    ...
	0xB000,		// 1011 0000 0000 0000b  ( 84)  T    -
	0x5B00,		// 0101 1011 0000 0000b  ( 85)  U    ..-
	0x56C0,		// 0101 0110 1100 0000b  ( 86)  V    ...-
	0x6B00,		// 0110 1011 0000 0000b  ( 87)  W    .--
	0x96C0,		// 1001 0110 1100 0000b  ( 88)  X    -..-
	0x9AC0,		// 1001 1010 1100 0000b  ( 89)  Y    -.--
	0xA5C0		// 1010 0101 1100 0000b  ( 90)  Z    --..
};
/*}}}  */


/*{{{  Interrupt Routine*/
/* ------------------------------------------------------------------------
 * Interrupt Routine
 * 250 cycle interrupt interval.
 * .25us * 250 cycles * 16 samples = 1000.0Hz frequency tone at 31.25 baud
 * ------------------------------------------------------------------------
 */

unsigned char DACvalue;

void interrupt tc_int(void) {
    TMR0 = 6+17;
    switch(ModType) {
        case RISING:  DACvalue = Waveform[(ModCount&0x07)+(((~ModCount)>>1)&0x78)];  break;
        case FALLING: DACvalue = Waveform[(ModCount&0x07)+((ModCount>>1)&0x78)];     break;
        case FLAT:    DACvalue = Waveform[ModCount&7];                               break;
        case NONE:    DACvalue = 128; break;
    };

    if(ModCount&8) DACvalue=~DACvalue; // Create negative if needed
    if(ModPhase) DACvalue=~DACvalue;   // Invert sample if phase is 180 degre.
    PORTB = DACvalue;
    ModCount+=1;

    T0IF = 0;
}
/*}}}  */


/*{{{  eeprom function*/

void EEPromWrite(addr, value) {
    while(WR)continue;
    EEADR=(addr);
    EEDATA=(value);
    GIE=0;
    WREN=1;
    EECON2=0x55;
    EECON2=0xAA;
    WR=1;
    WREN=0;
    GIE=1;
}

unsigned char EEPromRead(unsigned char addr) {
    return ((EEADR=(addr)),(RD=1),EEDATA);
}
/*}}}  */

/*{{{  Serial port function*/

#define baud 31            // Bit delay for 2400 baud (4-MHz PIC clock).
#define ser_o       7,6    // Serial Output Pin RC6
#define ser_i       7,7    // Serial Input Pin RC7

byte temp;	               // Temporary Register (used by serial,)
byte bits;	               // Bit counter used with Serial Routines
byte xbyte;                // Transmit/Receive Register for Serial Routines

/*{{{  sbit   - basic delay function*/

#asm
;-------------------------------------------------------------------------
; Low-Level Serial Routine  (FOR DEVICE WITHOUT USART !)
;
; Use: temp, xbyte, bits, (baud, sio_p)
;-------------------------------------------------------------------------

; --- Subroutine used by Serout to send a bit and generate time delays.
sBit_
        BTFSS 3,0             ; Output Carry flag on serial pin
        BCF ser_o
        BTFSC 3,0
        BSF ser_o

        MOVLW baud          ; Baud-rate delay in temp variable
        MOVWF _temp

	                        ; Delay routine used by Serin & Serout. Load the length of
	                        ; delay required into temp before calling.

sloop_
;        GOTO    sl1		    ; Two-cycle nops for time delay.
;sl1
;        GOTO    sl2		    ; "
;sl2
;        GOTO    sl3		    ; "
;sl3
;        GOTO    sl4		    ; "
;sl4
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP

        NOP			        ; Two-cycle nops = ( NOP+ CLRWDT )
    	CLRWDT		        ; Clear Watchdog !
        DECFSZ  _temp       ; Number of trips through loop
        GOTO    sloop_
    				        ; Set by baud rate in temp.
    	RETURN		        ; Return to calling routine.
#endasm
/*}}}  */

/*{{{  Serout - Serial output*/

void Serout(char car) {
    xbyte = car;
#asm
; -------------------------------------------------------------------------
; --- Transmits serial a single byte with a fixed output pin, baud rate,
; and polarity. Serout uses the common data format of no parity, 8 data
; bits, 1 stop bit (N81).
; - Upon entry, the desired pin must already be set up as an output.
; - Upon return, the data in the buffer will be unchanged

        BCF     11,7        ; Disable interrupt

        BCF     3,0	        ; Clear carry (Set up start bit, send Raw Data)
    	CALL	sBit_		; Send start bit.
        MOVLW   8           ; Send 8 data bits.
        MOVWF   _bits

xbit_
        RRF	    _xbyte  	; Rotate bit into carry.
    	CALL	sBit_		; Send the data bit.
        DECFSZ  _bits       ; Number of trips through loop
        GOTO    xbit_
    	RRF    _xbyte	    ; Rotate data back into original position.
    	BSF     3,0		    ; Set up stop bit.
    	CALL	sBit_		; Send the stop bit.

        BSF     11,7        ; Enable interrupt
#endasm
}
/*}}}  */

/*{{{  Serin  - Serial input (temp=255 if caracter received else temp=0) */

unsigned char Serin() {
#asm
;-------------------------------------------------------------------------
;--- Receive serially a single byte with a fixed input pin, baud rate,
; and polarity. Serin uses the common data format of no parity, 8 data
; bits, 1 stop bit (N81).
; - Upon entry, the desired pin must already be set up as an input.
; - Upon return, the data in the buffer wil be the received byte.
;
;   IF TEMP=0 NO CARACTER RECEIVED -> ELSE TEMP=255.
;

        CLRF    _xbyte		 ; Clear Receive Register

rpoll_
        CLRF    _temp
        CLRWDT  		     ; Clear Watchdog
        BTFSC   ser_i
        RETLW   0

        BCF     11,7        ; Disable interrupt

        MOVLW   baud/2       ; Wait 1/2 bit time to middle of start bit.
        MOVWF   _temp
        CALL	sloop_		 ; Call Bit Delay Sub-Routine
        BTFSC   ser_i
        GOTO    rpoll_

        MOVLW   8            ; Load 8 bit into xbyte
        MOVWF   _bits
rcv_
        MOVLW   baud         ; Wait bit time.
        MOVWF   _temp
        CALL	sloop_		 ; Call Bit Delay Sub-Routine
        BTFSS   ser_i        ; Put serial pin into carry
        BCF     3,0
        BTFSC   ser_i
        BSF     3,0
        RRF     _xbyte	 	 ; Rotate carry into msb of data byte.
        DECFSZ  _bits        ; Bits=bits-1. If bits<>0 then rcv_.
        GOTO    rcv_

        MOVLW   baud         ; Delay 1 bit time for stop bit.
        MOVWF   _temp
        CALL	sloop_		 ; Call Bit Delay Sub-Routine
        DECF    _temp

        BSF     11,7         ; Enable interrupt
#endasm
    return xbyte;
}
/*}}}  */

/*{{{  SerStr - send string to serial port*/

void SerStr(const char *p) {
    unsigned char i;

    i=0;
    while(1) {
       if(p[i]==0) return;
       Serout(p[i++]);
    }
}
/*}}}  */

/*{{{  SerHex - Send 8-bit hex number to serial port*/

void SerHex(unsigned char c) {
    Serout(hex[c>>4]);
    Serout(hex[c&15]);
}
/*}}}  */

/*}}}  */

/*{{{  PSK function*/

/*{{{  PSKStartSync - Send 60 zero*/
//-----------------------------------------------
// Send 60 zero (phase reversal)
//-----------------------------------------------
void PSKStartSync() {
    unsigned char i;

    for(i=0; i<60; i++) {
        WaitCycleEnd();
        ModType=FALLING;
        WaitCycleStart();

        WaitCycleEnd();
        if(ModPhase)          // Invert phase in middle of cycle
            ModPhase=0;
         else
            ModPhase=1;
        ModType=RISING;
        WaitCycleStart();
    }
};
/*}}}  */

/*{{{  PSKPutc - Send caracter to PSK31*/

void PSKPutc(unsigned char c) {
    unsigned int code;
    unsigned char zero;

    code = Varicode[c&127];
    zero = 0;

    while(zero!=2) {
        if(code&0x8000) {
            /*{{{  Constant phase*/

    // Set modulation FLAT for all cycle
    WaitCycleEnd();
    ModType=FLAT;
    WaitCycleStart();
    WaitCycleEnd();
    WaitCycleStart();
/*}}}  */

            zero=0;
        } else {
            /*{{{  Reverse phase*/

    WaitCycleEnd();
    ModType=FALLING;
    WaitCycleStart();

    WaitCycleEnd();
    if(ModPhase)          // Invert phase in middle cycle
        ModPhase=0;
     else
        ModPhase=1;
    ModType=RISING;
    WaitCycleStart();
/*}}}  */

            zero++;
        }
        code = code<<1;
    }
}
/*}}}  */

/*{{{  CWPutc - Send caracter in CW*/
////////////////////  Various constant tabels  //////////////////////////
// The word in the CW table is divided into 8 groups of two bits starting
// at the msb side.  The two bits represent one of four possible states.
//  00 - end of character
//  01 - DOT
//  10 - DASH
//  11 - SPACE of two dot times

void Dot() {
    unsigned char i;

    for(i=0; i<5; i++) {
        WaitCycleEnd();
        WaitCycleStart();
    }
};

void CWPutc(unsigned char c) {
    unsigned int code;

    if(c<32 || c>90) return;
    code = morse[c-32];

    while(1) {
        if((code&0xC000)==0xC000) Dot();
        if((code&0xC000)==0x8000) { ModType=FLAT; KEY=1; Dot(); Dot(); Dot(); KEY=0; }
        if((code&0xC000)==0x4000) { ModType=FLAT; KEY=1; Dot(); KEY=0; }
        if((code&0xC000)==0x0000) break;
        ModType=NONE; Dot();
        code=code<<2;
    }
}
/*}}}  */

/*}}}  */

/*{{{  Misc function*/

/*{{{  Hex2Dec - Conversion function*/

//-------------------------------------------
// Convert binary to hexadecimal caracter
//-------------------------------------------
unsigned char Hex2Dec(char c) {
    if(c>='0' && c<='9') return c-48;
    if(c>='A' && c<='F') return c-55;
    if(c>='a' && c<='f') return c-87;
    return 255;
}
/*}}}  */

/*{{{  WaitSec - Wait x second*/
//-------------------------------------------
// Wait aprox. 1 sec (0.992sec)
//-------------------------------------------
void WaitSec(unsigned char c) {
    unsigned int i;

    for(i=0; i<(unsigned int)c*62; i++) {      // wait 0.992sec
        WaitCycleEnd();
        WaitCycleStart();
        if(SERIN==0) return;
    }
};
/*}}}  */


/*}}}  */


/*{{{  Main*/


void main() {
    byte c,i,mode;

    TRISA   = 0b11111111;       // All input
	TRISC   = 0b10001111;       // serial port in/out RC6+RC7 PTT and KEY output
	TRISB   = 0b00000000;       // R-R2 DAC output port

    OPTION  = 0b00001111;       // Pull-up + WDT prescale at 1:128
	INTCON  = 0b00100000;       // TMR0 interrupt
	PIE1    = 0b00000000;       // No peripheral interrupt
	PIE2    = 0b00000000;
    T1CON   = 0b00000000;       // TMR1 is off
    T2CON   = 0b00000000;       // TMR2 is off
    CCP1CON = 0b00000000;       // Capture/compare/pwm #1 is off
    SSPCON  = 0b00000000;       // Sync serial port is off
    ADCON0  = 0b00000000;       // A/D is off
    ADCON1  = 0b00000110;       // A/D is off

    TMR0 = 6+14;                // For 16 sample at 1khz@16Mhz
                                // 1000us = 1khz period, Tcycle=0.25us@20Mhz
                                // (1000us/16)/Tcycle
                                // 62.5us / 0.25us = 250 cycles between interrupt
                                // 256 - 250 = 6 + 14(delay to reload timer)
                                // 250 cycles * 0.25us * 16 samples = 1000.0hz at 31.25baud

    ModType = FLAT;
    PTT = 0;
    KEY = 0;
    SerStr("PSK31 beacon for PIC, by VE2YAG\r\r");

    T0IE = 1;           // Enable TMR0 interrupt
    while(1) {
        /*{{{  Sending mode*/

    i = 0;
    mode = MODE_PSK;

    while(1) {

        if(SERIN==0) break;       // Check if key pressed

        //------------------------------------------
        //  Verify for end of message, back to PSK,
        //  stop carrier and wait delay
        //------------------------------------------
        if(i==EEPromRead(EE_BEACONTEXTSIZE)) {
            ModType = NONE;
            mode = MODE_PSK;
            PTT=0;
            i=0;
            WaitSec(EEPromRead(EE_BEACONDELAY));
            continue;
        }

        c = EEPromRead(EE_BEACONTEXT+i);    // Read caracter to send

        //-------------------------------
        //  Switch to CW, stop carrier
        //-------------------------------
        if(c=='^') { ModType = NONE;
                     mode = MODE_CW;
                     i++;
                     continue;
                   };

        PTT=1;
        i++;

        //-------------------------------
        //  TONE sending
        //-------------------------------
        if(c=='~') {
            ModType = FLAT;
            WaitSec(1);
            continue;
        }

        //-------------------------------
        //  PSK sending
        //-------------------------------
        if(mode == MODE_PSK) {
            if(i==1) PSKStartSync();
            PSKPutc(c);
            continue;
        }

        //---------------------------------------------
        //  CW sending
        //---------------------------------------------
        CWPutc(toupper(c));

    }
/*}}}  */

        /*{{{  Config mode*/

//-----------------------------
// Enter in config mode
//-----------------------------
    ModType = NONE;
    PTT = 0; KEY=0;
    LineCount = 0;
    T0IE = 0;   // Disable TMR0 interrupt

    SerStr("Config mode: ? for help\r>");

    while(1) {
        c = Serin();
        if(temp) {
            if( (c==DEL || c==BK) && LineCount) { SerStr(bk_str); LineCount--; continue; }
            if(c==CR) { /*{{{  Process command*/
    
    if(LineCount==0) { SerStr("\r>"); continue; }

    switch(LineBuffer[0]) {
        case 'R': /*{{{  R command*/

 SerStr("\rBeacon Activated\r"); LineCount=0; /*}}}  */
 break;
        case 'B': /*{{{  B command*/

    for(c=0; c<LineCount-1; c++) EEPromWrite(EE_BEACONTEXT+c, LineBuffer[c+1]);
    EEPromWrite(EE_BEACONTEXTSIZE, LineCount-1);
/*}}}  */
 break;
        case 'E': /*{{{  E command*/

    c = Hex2Dec(LineBuffer[1]);
    temp = Hex2Dec(LineBuffer[2]);

    if((c==255) || (temp==255)) { SerStr("\r?error"); break; }

    EEPromWrite(EE_BEACONDELAY, (c*16)+temp);
/*}}}  */
 break;
        case 'D': /*{{{  D command*/

    SerStr("\r!");
    SerHex(CFG_CPUTYPE);
    SerHex(CFG_SOFTVERSION);
    SerHex(CFG_EEPROMSIZE);
    SerHex(0x00);

    for(i=0; i<CFG_EEPROMSIZE; i++) SerHex(EEPromRead(i));
/*}}}  */
 break;
        case '?': /*{{{  ? command*/

    SerStr("\rB<text> ~ for 1sec tone ^ to switch CW\rExx     (Second to wait in hex)\rR       (Restart)\rD       (eeprom hex dump)");
/*}}}  */
 break;
        default:  SerStr("\r?error");
    }

    if(LineCount==0) break; // Reset

    SerStr("\r>");
    LineCount=0;
/*}}}  */
 }
            if(LineCount<MAXCARACTER && c>31 && c<127) {
                LineBuffer[LineCount++]=c;
                Serout(c);
            }
        }
    }

    T0IE = 1;   // Enable TMR0 interrupt
/*}}}  */

    }
}

/*}}}  */

